" Vim syntax file
" Language:	test
" Maintainer:	Mel Flynn <rflynn@acsalaska.net>
" URL:		http://bitbucket.org/bits/vim/pmake.vim
" Last Change:	2012 Apr 21
"
" $Coar$
"
" vim: tw=158
" Hi! I'm using a 160 column vesa screen. It's awesome. Also, it makes these files a lot more readable and maintainable. Please file your complaints with our
" complaints manager, Mr. Devon Null <devnull@wecare.com>.

" For version 5.x: Clear all syntax items
" For version 6.x: Quit when a syntax file was already loaded
if version < 600
  syntax clear
elseif exists("b:current_syntax")
  finish
endif

"
" GENERAL ISSUES
"
" The difference between a dependency a target:
" - A dependency always has a source
" - A target looks like a dependency, but doesn't have to have a source and also it is followed by at least one shell command
" - Followed means:
"   - after stripping comments
"   - after ignoring statements and loops
"   - before any empty line or start of new target or dependency
" - If the above is true, it is not a dependency but a target until the first empty line or the start of a new target or dependency.
" - If the above is not true, anything after a dependency is up for grabs again.
"
" Given the above we don't identify dependency lines or - if you prefer the PR department's answer - we color dependencies equal to targets because they are
" very similar in nature.
"
" Special sources and targets, as well as special variables contain a dot. Shell keywords within test contain a leading dash.
setlocal iskeyword+=.
setlocal iskeyword+=-

" Line ending stuff
syn match pmakeNextLine		/\\$/ contained
syn match pmakeEmptyLine	/^$/
syn match pmakeEndOfLine	/$/ contained

"
" VARIABLES
"
" Assignment
syn match   pmakeVarAssign	/^[_$A-Za-z][a-zA-Z0-9_${}()]\+=/me=e-1 nextgroup=pmakeVarAssignType contains=pmakeVarExpansion
syn match   pmakeVarAssign	/^[_$A-Za-z][a-zA-Z0-9_${}()]\+[?!+:]=/me=e-2 nextgroup=pmakeVarAssignType contains=pmakeVarExpansion
syn match   pmakeVarAssignType	/[?!+:]\?=/ contained nextgroup=pmakeVarValues skipwhite
syn region  pmakeVarValues	start=/./ end=/$/ contained contains=pmakeNextLine,pmakeVarExpansion,pmakeComment
" Expansion
syn region  pmakeVarExpansion	excludenl start=/\${/ end=/}/ contained contains=pmakeVarExpansion,pmakeSpecialVar,pmakeVarModifier,pmakeVarModError
syn match   pmakeVarModifier	/:[EHLOQRTUu][:}]/me=e-1 contained
" Known working regex separators: / @ , ^ | %
" Known highlighting failure (from bsd.port.mk):
"
" _S_TEMP=	${_S:S/^${_S:C@/?:[^/:]+$@/@}//:S/^://}
"
" However, it can be fixed like so:
" _S_TEMP=	${_S:S/^${_S:C@/?:[^:/]+$@/@}//:S/^://}
"                                   ^^
" Thus, I'm not enclined to complicate this any further. I mean we blame vim syntax highlighting because you cannot assign it regions in which everything you
" said it could contain and everything you told the thing that this thing is contained in could contain and whereever that may be contained in could contain
" and so on should no longer apply. Punctuation in the previous sentence has been omitted on purpose.
syn region  pmakeVarModifier	start=/:[SC]\// skip=/\\\// end=/\/g\?[:}]/me=e-1 contained contains=pmakeVMRegex keepend
syn region  pmakeVarModifier	start=/:[SC]@/ skip=/\\@/ end=/@g\?[:}]/me=e-1 contained contains=pmakeVMRegex keepend
syn region  pmakeVarModifier	start=/:[SC],/ skip=/\\,/ end=/,g\?[:}]/me=e-1 contained contains=pmakeVMRegex keepend
syn region  pmakeVarModifier	excludenl start=/:[MN]/ end=/[:}]/me=s-1 contained contains=pmakeVMGlob
syn match   pmakeVarModError	/:[^EHLOQRTUuCS][:}]/me=e-1 contained
syn match   pmakeVarModError	/:[CS][^@^,%|/]/ contained
syn match   pmakeVarModError	/:[MN]}/ contained
syn region  pmakeVMGlob		start=/:[MN]/ms=e+1 skip=/\\:/ end=/[:}]/me=s-1 contained contains=pmakeVarExpansion
syn region  pmakeVMRegex	start=/\// skip=/\\\// end=/\/g?[:}]/me=s contained contains=pmakeVarExpansion
syn region  pmakeVMRegex	start=/@/ skip=/\\@/ end=/@g?[:}]/me=s contained contains=pmakeVarExpansion
syn region  pmakeVMRegex	start=/,/ skip=/\\,/ end=/,g?[:}]/me=s contained contains=pmakeVarExpansion
syn region  pmakeVMRegex	start=/\^/ skip=/\\\^/ end=/\^g?[:}]/me=s contained contains=pmakeVarExpansion
syn region  pmakeVMRegex	start=/|/ skip=/\\|/ end=/|g?[:}]/me=s contained contains=pmakeVarExpansion
syn region  pmakeVMRegex	start=/%/ skip=/\\%/ end=/%g?[:}]/me=s contained contains=pmakeVarExpansion
syn keyword pmakeSpecialVar	.CURDIR .TARGET .OBJDIR .IMPSRC .ALLSRC contained
"
" FUNCTIONS
"
" Functions can only exist in pmakeCondIf parts. They start with the function name and are followed by parenthesis. They can enclose different things based
" on the function.
"
" Defined and empty contain the same variable identifier
syn match   pmakeFuncDefined	/\(defined\|empty\)\>/ nextgroup=pmakeFuncDefinedPar skipwhite contained
syn match   pmakeFuncDefinedPar	/([^)]\+)/ contained contains=pmakeVarName,pmakeVarExpansion
syn match   pmakeVarName	/[a-zA-Z_.][a-zA-Z0-9_]\+/ contained
" Exists may only contain a path
syn match   pmakeFuncExists	/exists\>/ nextgroup=pmakeFuncExistsPar skipwhite contained
syn match   pmakeFuncExistsPar	/([^)]\+)/ contained contains=pmakePath,pmakeVarExpansion
syn match   pmakePath		/[/.a-zA-Z0-9_ \\-]\+/ contained
" Target may only contain target names
syn match   pmakeFuncTarget	/target\>/ nextgroup=pmakeFuncTargetPar skipwhite contained
syn match   pmakeFuncTargetPar	/([^)]\+)/ contained contains=pmakeTarget,pmakeVarExpansion
syn match   pmakeTarget		/[a-zA-Z./][a-zA-Z0-9-]\+/ contained
"
" STATEMENTS
"
" Conditions
" 
" - In an earlier incarnation we tried to define regions and be smart about what can be contained inside condition statements. This turned out to give great
"   problems and in the end we also had to conclude that everything is allowed that is also allowed in the global space. Hence, we now stick to line-based
"   parsing and do not care about the lines that follow.
" - The drawback of the above is that we cannot identify mis-matched if/endif clauses. However, vim is not really able to do this to begin with as described
"   in 44.10 of the user manual:
"     Compilers have it easy.  They start at the beginning of a file and parse it
"     straight through.  Vim does not have it so easy.  It must start in the middle,
"     where the editing is being done.
"   As such, it would generate a lot of false positive errors (ya know, parts highlighted as errors that are not errors because they are false positives -
"   this is an error), especially when large parts are enclosed in if defined(FOO) statements.
"
" First, the easy ones: endif and else
syn match   pmakeCondEndIf	/^\.\s*\(endif\|else\)\>/ nextgroup=pmakeComment,pmakeEndOfLine skipwhite
" Next, ifmake and ifdef can only be followed by IDENTIFIER and do not allow combinations with other conditions.
syn match   pmakeCondIfdef	/^\.\s*\(el\)\?ifn\?def[\\t ]\+/ nextgroup=pmakeCondVarName skipwhite
syn match   pmakeCondIfmake	/^\.\s*\(el\)\?ifn\?make[\\t ]\+/ nextgroup=pmakeCondTarName skipwhite
syn match   pmakeCondVarName	/[a-zA-Z_.][a-zA-Z0-9_]\+/ contained contains=pmakeSpecialVar nextgroup=pmakeComment,pmakeEndOfLine
syn match   pmakeCondTarName	/[a-zA-Z.][a-zA-Z0-9-]\+/ contained contains=pmakeSpecialTargets nextgroup=pmakeComment,pmakeEndOfLine
" Finally, if and elif:
" The basics is:
"   - . at start of line
"   - whitespace or the absence thereof (blackspace or whitevoid?)
"   - if or elif
"   - end of word
"   - match till end of line
"   - skip line continuations
" The problem is what's after the if or elif.
syn region  pmakeCondIf		start=/^\.\s*\(el\)\?if\>/ end=/$/ contains=pmakeNextLine,@pmakeConditions,pmakeComment,pmakeCondNegate,pmakeCondGrouping
syn match   pmakeCondNegate	/!/ contained
syn match   pmakeCondLogical	/\(&&\|||\)/ contained
syn match   pmakeCondCompare	/\(==\|!=\)/ contained nextgroup=pmakeCondString,pmakeVarExpansion skipwhite
syn match   pmakeCondString	/\S\+/ contained
syn region  pmakeCondString	start=/"/ end=/"/ contained
syn region  pmakeCondGrouping	matchgroup=pmakeParen start=/(/ end=/)/ contained contains=@pmakeConditions,pmakeNextLine,pmakeCondNegate
syn cluster pmakeConditions	contains=pmakeFuncDefined,pmakeFuncExists,pmakeFuncTarget,pmakeCondLogical,pmakeCondCompare,pmakeVarExpansion
" Loops (or better: `The for loop')
"
" We define the for loop as a PreProc. For loops are parsed, unwound and it's iterator variable expanded before anything else. While it may in practice look
" like a normal loop it is not. You will experience this when trying to use the iterator as a variable in comparisons.
"
" As for the syntax, it's rather simple: we only identify the .for and .endfor line. Anything in between is not our business.
"
" Let's start with endfor:
syn match   pmakePPEndFor	/^\.\s*endfor\>/ nextgroup=pmakeComment,pmakeEndOfLine skipwhite
" Now the for loop itself
syn match   pmakePPFor		/^\.\s*for\>/ nextgroup=pmakePPForIterator skipwhite
syn match   pmakePPForIterator	/[a-zA-Z0-9_]\+/ contained nextgroup=pmakePPForIn skipwhite
syn match   pmakePPForIn	/in\>/ nextgroup=pmakePPForList contained skipwhite
syn region  pmakePPForList	start=/\S/ end=/$/ contained contains=pmakeNextLine,pmakeVarExpansion,pmakeComment
" undef: the odd ball
"
" Almost forgot about undef. It undefines variables. At first glance, it seems like a useless operation. In many cases, it is. It is far better not to define
" the variable in the first place. The culprit however, is our entry at `finally'.
syn match   pmakeUndef		/^\.\s\*undef\>/ nextgroup=pmakeUndefVarname
syn match   pmakeUndefVarname	/[a-zA-Z0-9_.][a-zA-Z0-9_]\+/ nextgroup=pmakeEndOfLine,pmakeComment contained
"
" And finally: the include
"
" Alongside a normal include Pmake has a silent include. It's one of those things that sneaks in and you don't really notice until it's not there. Then things
" go terribly wrong. Also, silent is a bad name choice. The normal include is actually quite silent. When it can't find the guest to invite, it politely tells
" you so and it shuts up the entire make program for you too. Now, the so called `silent include' will keep it's mouth shut if he can't find the guest to
" invite to the party, however, if the rest of the party counted on that guest bringing the beer ... well ... they all start screaming very loudly!
"
" As for syntax: the keyword followed by an argument that shall be resolved to a filename. In fact, when enclosed in brackets, we only allow filenames. We do
" not allow paths.
syn match   pmakeInclude	/^\.\s*s\?include\>/ nextgroup=pmakeIncludeArg skipwhite
syn match   pmakeIncludeArg	/<[a-zA-Z0-9_.-]\+>/ nextgroup=pmakeComment,pmakeEndOfLine contained
syn match   pmakeIncludeArg	/"[a-zA-Z0-9_./]\+"/ nextgroup=pmakeComment,pmakeEndOfLine contained contains=pmakeVarExpansion
"
" TARGETS
"
" These are the Iksar of syntax highlighting. KoS everywhere, yet somehow they show up around each corner.
"
" Let us start simple.
syn keyword pmakeSpecialTargets	.BEGIN .DEFAULT .END .IGNORE .INCLUDES .INTERRUPT .LIBS .MAIN .MAKEFILEDEPS .MAKEFLAGS .MFLAGS .NOTPARALLEL contained
syn keyword pmakeSpecialTargets	.ORDER .PATH .PHONY .PRECIOUS .SHELL .SILENT .SUFFIXES .WARN contained
syn keyword pmakeSpecialSources	.IGNORE .MAKE .NOTMAIN .OPTIONAL .PRECIOUS .SILENT .USE .WAIT contained
" Next we try to identify a target definition
" - the easy one: foo bar baz *.glob:\n
syn region  pmakeTargetDef	oneline start=/^/ matchgroup=pmakeTargetEOT end=/::\?$/ contains=pmakeSpecialTargets,pmakeVarExpansion
" - target with sources:
"   We cheat a little, to allow variable assignments like so:
"     PUNKIGNORE= version mistmatch: need foo!!!! you have bar :(((( were u mine 4 fish?!?!??!
"   Then again, we're not entirely sure if a target definition can have equal signs.
syn match   pmakeTargetDef	/^[^=:]\+::\?\s/me=e-1 nextgroup=pmakeSources contains=pmakeVarExpansion,pmakeSpecialTargets skipwhite
syn region  pmakeSources	start=/\S/ end=/$/ contains=pmakeSpecialSources,pmakeVarExpansion,pmakeNextLine,pmakeComment contained
syn match   pmakeTargetEOT	/:/ contained
" - and finally, the target shell commands
"   For a brief moment I thought about including the sh.vim syntax file. Then I saw that it made wrong assumptions about FreeBSD's sh(1). I also saw it was
"   about twice as small as we were at the time. Also, it would do the wrong thing with respect to shell variables and pmake variables.
"   This is also where the descision to do line-based parsing only may at some point start to create errors, as all we have to identify a shell line is that
"   it starts with a tab. This is also true for continuations of for example variable assignments and if statements.
"   The good part is that these parts do not start with a tab and support line continuations, so in general I think we're good.
syn region  pmakeTargetShell	start=/^\t/ end=/$/ contains=pmakeNextLine,pmakeVarExpansion,pmakeComment,pmakeTSSilent,pmakeTSErrOk,@pmakeShellStuffs
syn cluster pmakeShellStuffs	contains=pmakeShellVar,pmakeShellKeywords,pmakeShellNextLine,pmakeShellTest,pmakeShellBackticks
syn region  pmakeShellVar	excludenl start=/\$\${/ms=s+1 end=/}/ contained
syn region  pmakeShellVar	excludenl start=/\$\$(/ms=s+1 end=/)/ contained
syn match   pmakeShellVar       /\$\$[a-zA-Z_][a-zA-Z0-9_]\+/ms=s+1
syn keyword pmakeShellKeywords  if then else elif fi set do done test for while until read echo case esac cd eval contained
syn region  pmakeShellTest	excludenl start=/\[/ end=/\]/ contained contains=pmakeShellTestKw,pmakeShellTestOp,pmakeVarExpansion,pmakeShellString,pmakeShellVar
syn region  pmakeShellString    start=/"/ end=/"/ contained contains=pmakeVarExpansion,pmakeShellVar
syn keyword pmakeShellTestKw	-eq -lt -gt -ne -a -o -s -f -e -r -x -z contained
syn match   pmakeShellTestOp	/\(!\|!=\|=\)/ contained
syn match   pmakeShellNextLine	/;/ contained
syn region  pmakeShellBackticks	matchgroup=pmakeShellBTticks start=/`/ end=/`/ contained contains=@pmakeShellStuffs,pmakeVarExpansion,pmakeNextLine
syn match   pmakeTSSilent	/^\t-\?@/ contained
syn match   pmakeTSErrOk	/^\t-[^@]/me=e-1 contained
" COMMENTS
"
" - Comments that don't start on a line can only be part of certain structures. Most notably they cannot be part of variable expansion.
syn match   pmakeComment	excludenl /#.*$/ contained
" - Comments cannot expand beyond a line, not even with \\ (contrary to popular belief).
syn region  pmakeCommentLine	start=/^#/ end=/$/ oneline

" Set the perty colors
hi link pmakeNextLine		Delimiter
hi link pmakeVarAssign		Identifier
hi link pmakeVarAssignType	Delimiter
hi link pmakeVarValues		String
hi link pmakeVarExpansion	Identifier
hi link pmakeVarModifier	Macro
hi link pmakeVarModError	Error
hi link pmakeCondIfdef		Conditional
hi link pmakeCondIfmake		Conditional
hi link pmakeCondEndIf		Conditional
hi link pmakeCondIf		Conditional
hi link pmakeCondNegate		Special
hi link pmakeCondLogical	Operator
hi link pmakeCondCompare	Operator
hi link pmakeCondGrouping	Delimiter
hi link pmakeCondString		String
hi link pmakeFuncDefined	Function
hi link pmakeFuncDefinedPar	Delimiter
hi link pmakeVarName		Identifier
hi link pmakeSpecialVar		Keyword
hi link pmakeVMRegex		Special
hi link pmakeVMGlob		String
hi link pmakeFuncExists		Function
hi link pmakeFuncExistsPar	Delimiter
hi link pmakePath		String
hi link pmakeFuncTarget		Function
hi link pmakeFuncTargetPar	Delimiter
hi link pmakePPFor		PreProc
hi link pmakePPForIterator	Identifier
hi link pmakePPForIn		PreProc
hi link pmakePPForList		String
hi link pmakePPEndFor		PreProc
hi link pmakeTarget		Type
hi link pmakeTargetDef		Type
hi link pmakeSpecialTargets	Keyword
hi link pmakeTargetEOT		Delimiter
hi link pmakeSources		Constant
hi link pmakeSpecialSources	Keyword
hi link pmakeTargetShell	String
hi link pmakeShellVar		PreProc
hi link pmakeShellKeywords	Statement
hi link pmakeShellNextLine	Delimiter
hi link pmakeShellTest		Statement
hi link pmakeShellTestKw	Operator
hi link pmakeShellTestOp	Operator
hi link pmakeShellString	String
hi link pmakeShellBackticks	Function
hi link pmakeShellBTticks	Special
hi link pmakeTSSilent		Special
hi link pmakeTSErrOk		Special
hi link pmakeCondVarName	Identifier
hi link pmakeCondTarName	Type
hi link pmakeParen		Delimiter
hi link pmakeComment		Comment
hi link pmakeCommentLine	Comment
hi link pmakeInclude		Include
hi link pmakeIncludeArg		String

let b:current_syntax = "pmake"
